# oscillode
# Copyright (C) 2024 oscillode Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Operator allowing the user to use the conventional CTRL+SHIFT+LMB to connect arbitrary node outputs to either an existing, or a new, viewer node."""

import bpy

from oscillode import contracts as ct
from oscillode.utils import logger

log = logger.get(__name__)


class ConnectViewerNode(bpy.types.Operator):
	"""Connecting the viewer node to the active node in a Maxwell Sim node tree, as defined by selection and mouse-cursor location.

	If there is no active viewer node, then one is created.
	"""

	bl_idname = ct.OperatorType.ConnectViewerNode
	bl_label = 'Connect Viewer to Active'
	bl_description = 'Connect active node to Viewer Node'
	bl_options = frozenset({'REGISTER', 'UNDO'})

	@classmethod
	def poll(cls, context: bpy.types.Context) -> bool:
		"""Require the active space to be a Maxwell Sim node tree defined by this addon.

		Notes:
			Run by Blender when determining whether the operator can run.

		Returns:
			Whether the operator can run.
		"""
		space = context.space_data
		return (
			space.type == 'NODE_EDITOR'
			and space.node_tree is not None
			and space.node_tree.bl_idname == ct.NodeTreeType.MaxwellSim
		)

	def invoke(
		self, context: bpy.types.Context, event: bpy.types.Event
	) -> ct.BLOperatorStatus:
		"""Require the active space to be a particular node tree defined by this addon.

		Notes:
			Run by Blender when determining whether the operator can run.

		Returns:
			Whether the operator can run.
		"""
		node_tree = context.space_data.node_tree

		# Retrieve Mouse Location
		mlocx = event.mouse_region_x
		mlocy = event.mouse_region_y

		# Select Node at Location
		## - Deselect all other nodes.
		bpy.ops.node.select(
			extend=False,
			location=(mlocx, mlocy),
		)
		selected_node = context.selected_nodes[0]

		# Find Viewer Node...
		viewer_node_type = node_tree.viewer_node_type
		for node in node_tree.nodes:
			# TODO: Support multiple viewer nodes.
			if hasattr(node, 'node_type') and node.node_type is viewer_node_type:
				viewer_node = node
				break

		# ...OR: Create Viewer Node
		else:
			# TODO: Place viewer where it doesn't overlap other nodes.
			viewer_node = node_tree.nodes.new(viewer_node_type)
			viewer_node.location.x = selected_node.location.x + 250
			viewer_node.location.y = selected_node.location.y
			selected_node.select = False

		# Remove Input Links from Viewer
		## - Unless the selected node is already linked to the viewer.
		should_create_new_link = True
		for link in viewer_node.inputs[0].links:
			if link.from_node.name == selected_node.name:
				should_create_new_link = False
				continue
			node_tree.links.remove(link)

		# Create Link from Selected Node to Viewer
		if should_create_new_link:
			node_tree.links.new(selected_node.outputs[0], viewer_node.inputs[0])

		return {'FINISHED'}


####################
# - Blender Registration
####################
BL_REGISTER: list[ct.BLClass] = [
	ConnectViewerNode,
]

BL_HANDLERS: ct.BLHandlers = ct.BLHandlers()

BL_KEYMAP_ITEMS: list[ct.BLKeymapItem] = [
	ct.BLKeymapItem(
		operator=ct.OperatorType.ConnectViewerNode,
		event_type='LEFTMOUSE',
		event_value='PRESS',
		ctrl=True,
		shift=True,
		space_type='NODE_EDITOR',
	)
]
